import * as fs from "fs";
import * as toml from "toml";
import FilePath from "./path";

export interface Metafile {
  name: string;
  filename: string;
  side: "server" | "client" | "both";
  provider: ModrinthProvider | CurseForgeProvider;
}

export interface ModrinthProvider {
  url: string;
  hashFormat: string;
  hash: string;
  modId: string;
  versionId: string;
}

export interface CurseForgeProvider {
  hashFormat: string;
  hash: string;
  mode: string;
  fileId: number;
  projectId: number;
}

/**
 * Represents a file entry in the Packwiz index.
 */
export class IndexFileEntry {
  readonly file: FilePath;
  readonly hash: string;
  readonly metafile: boolean;

  constructor(
    file: string | FilePath,
    hash: string,
    metafile: boolean = false,
  ) {
    this.file = file instanceof FilePath ? file : new FilePath(file);
    this.hash = hash;
    this.metafile = metafile;
  }

  parse(): Metafile {
    if (this.metafile === false) {
      throw new Error(
        "Cannot parse non-metafiles! Use `file.readText()` instead to get the file contents.",
      );
    }
    const fileContent = this.file.readText();
    const parsed = toml.parse(fileContent);

    const metafile: Metafile = {
      name: parsed.name,
      filename: parsed.filename,
      side: parsed.side,
      provider: this.parseProvider(parsed),
    };

    return metafile;
  }

  private parseProvider(parsed: any): ModrinthProvider | CurseForgeProvider {
    if (parsed.update?.modrinth) {
      return {
        url: parsed.download.url,
        hashFormat: parsed.download["hash-format"],
        hash: parsed.download.hash,
        modId: parsed.update.modrinth["mod-id"],
        versionId: parsed.update.modrinth["version"],
      };
    } else if (parsed.update?.curseforge) {
      return {
        hashFormat: parsed.download["hash-format"],
        hash: parsed.download.hash,
        mode: parsed.download.mode,
        fileId: parsed.update.curseforge["file-id"],
        projectId: parsed.update.curseforge["project-id"],
      };
    } else {
      throw new Error("Unknown provider in TOML.");
    }
  }
}

/**
 * Represents the structure of a Packwiz index file, as well as providing its location.
 */
export interface PackwizIndex {
  location: FilePath;
  hashFormat: string;
  files: IndexFileEntry[];
}

// TODO: parse pack.toml too instead of just the index file
// This is because pack.toml contains important information like modloader versions

/**
 * Parses a Packwiz index file and returns its contents.
 *
 * @param indexFilePath - The path of the TOML file.
 * @returns The parsed index as a structured object.
 */
export function parsePackwizIndex(indexFilePath: string): PackwizIndex {
  const indexFile = new FilePath(indexFilePath);
  const parsed = toml.parse(indexFile.readText());
  return {
    location: indexFile,
    hashFormat: parsed["hash-format"],
    files: parsed.files.map(
      (file: any) =>
        new IndexFileEntry(
          indexFile.parent.join(file.file),
          file.hash,
          file.metafile,
        ),
    ),
  };
}
